package App::Beam::Backend::Tar;

use strict;
use Carp;

require App::Beam::Backend;
use Data::Dumper;
our @ISA = qw(App::Beam::Backend);

use POSIX qw(strftime);
use App::Beam;
use App::Beam::Command qw(:channels);
use App::Beam::History::Entry qw(:state);

sub ck_dir {
    my ($v) = @_;
    
    if (-d $$v) {
	return undef;
    } elsif (! -e $$v) {
	return "directory does not exist";
    } else {
	return "not a directory";
    }
}

my %synt = (
    backend => {
	section => {
	    tar => {
		section => {
		    binary => {
			default => '/bin/tar',
			check => sub {
			    my ($v) = @_;
			    if (! -x $$v) {
				if (! -e $$v) {
				    return "binary does not not exist";
				} else {
				    return "not an executable";
				}
			    }
			    return undef;
			}
		    },
		    options => {
			array => 1
		    },
		    suffix => { default => 'tar' },
		    'snapshot-dir' => {
			default => '/var/lib/backups',
			check => \&ck_dir,
		    },
		}
	    }
	}
    },
    item => {
	section => {
	    '*' => {
		select => sub {
		    my ($vref, @path) = @_;
		    return 0 unless ref($vref) eq 'HASH';
		    return $vref->{backend}->{-value} eq 'tar';
		},
		section => {
		    backend => 1,
		    directory => {
			mandatory => 1,
			check => \&ck_dir
		    },
		    files => 1,
		    options => {
			array => 1
		    }
		}
	    }
	}
    }
);

sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);

    unless ($self->{beam}->lint(\%synt)) {
	return undef;
    }

    return $self;
}

sub snapshot_name {
    my ($self, $basename, $idx) = @_;
    my $name = $self->{beam}->format_name($basename, $idx);
    if ($name) {
	$name = $self->get('backend', 'tar', 'snapshot-dir')
	        . '/' . $name . '.db';
    }
    return $name;
}

sub mksnapshot {
    my ($self, $item) = @_;
    my $snapshot = $self->snapshot_name($item);
    if ($self->record->level != 0) {
	my $prev = $self->snapshot_name($item, 1);
	if ($prev) {
	    $self->debug(2, "cp $prev $snapshot");
	    unless ($self->dry_run) {
		use File::Copy;
		unless (copy($prev, $snapshot)) {
		    $self->error("can't copy $prev to $snapshot: $!");
		} 
	    }
	} else {
	    $self->error("$item: warning: can't locate previous snapshot\n"
			 . "falling back to level 0");
	}
    }
    return $snapshot;
}

sub backup {
    my ($self, $name, $item) = @_;   

    my $basename = $self->{beam}->format_name($name);
    croak "undefined basename" unless defined $basename;
    my $archive = $self->get('core', 'archivedir')
                  . '/'
	          . $basename
		  . '.'
		  . $self->get('backend', 'tar', 'suffix');
    
    my $cmd = new App::Beam::Command($self->get('backend', 'tar', 'binary'));
    $cmd->set_logger(CHAN_STDERR, sub {	$self->logger('err', $_) });
    if ($self->isset('backend', 'tar', 'options')) {
	$cmd->add($self->get('backend', 'tar', 'options'));
    }
    if ($self->isset('item', $name, 'options')) {
	$cmd->add($self->get('item', $name, 'options'));
    }
    $cmd->add('-c');
    $cmd->add('-f', $archive);
    $cmd->add('-C', $self->get('item', $name, 'directory'));
    my $snar = $self->mksnapshot($name);
    $cmd->add('--listed', $snar);

    if ($self->isset('item', $name, 'files')) {
	$cmd->add($self->get('item', $name, 'files'));
    } else {
	$cmd->add('.');
    }

    $item->add($archive, $snar);

    $self->debug(2, "running ".$cmd->command_line);
    my $ret = STATE_SUCCESS;
    unless ($self->dry_run) {
	$cmd->run;
	$self->logcommand($cmd);
	my $rc = $cmd->exit_code;
	$ret = STATE_FAILURE unless ($rc == 0 || $rc == 2);
    }
    return $ret;
}

sub find_full_backup {
    my ($self, $index, $item) = @_;

    my $rec = $self->record($index);
    my $cycle = $rec->cycle;
    
    my $i;
    for ($i = $index; $rec = $self->record($i); $i++) {
	unless ($rec->cycle == $cycle) {
	    $self->logger('err', "can't find full backup of $item starting from $index");
	    last;
	}
	unless (defined($rec->state($item))) {
	    $self->logger('err', "no item $item in backup $i");
	    last;
	}
	return $i if $rec->level == 0;
    }
    return undef;
}

sub restore_from {
    my ($self, $index, $item) = @_;

    my $basename = $self->{beam}->format_name($item, $index);
    croak "undefined basename" unless defined $basename;
    my $archive = $self->get('core', 'archivedir')
                  . '/'
	          . $basename
		  . '.'
		  . $self->get('backend', 'tar', 'suffix');
    $self->debug(1, "restoring \"$item\" from $archive");

    my $cmd = new App::Beam::Command($self->get('backend', 'tar', 'binary'));
    $cmd->set_logger(CHAN_STDERR, sub {	$self->logger('err', $_) });
    if ($self->isset('backend', 'tar', 'options')) {
	$cmd->add($self->get('backend', 'tar', 'options'));
    }
    if ($self->isset('item', $item, 'options')) {
	$cmd->add($self->get('item', $item, 'options'));
    }
    $cmd->add('-x');
    $cmd->add('-f', $archive);
    $cmd->add('-C', $self->get('item', $item, 'directory'));
    $cmd->add('--listed', '/dev/null');

    $self->debug(2, "running ".$cmd->command_line);
    my $ret = STATE_SUCCESS;
    unless ($self->dry_run) {
	$cmd->run;
	$self->logcommand($cmd);
	my $rc = $cmd->exit_code;
	$ret = STATE_FAILURE unless ($rc == 0 || $rc == 2);
    }
    return $ret;    
}

sub restore {
    my ($self, $index, $item) = @_;
    my $i = $self->find_full_backup($index, $item);
    return unless defined $i;

    my $result = STATE_SUCCESS;
    
    while ($i >= $index) {
	if ($self->restore_from($i, $item) != STATE_SUCCESS) {
	    $result = STATE_FAILURE;
	}
	$i--;
    }
    return $result;
}

1;
