package App::Beam::List;

use strict;
use Carp;
use POSIX qw(strftime);
use Unix::Sysexits;
use Getopt::Long qw(GetOptionsFromArray);

use App::Beam;
our @ISA = qw(App::Beam);

use App::Beam::History;
use App::Beam::History::Record;
use App::Beam::History::Entry;

=head1 NAME

beam list - list existing backups

=head1 SYNOPSIS

B<beam>
[I<OPTIONS>]
B<list>
[B<-v>]
[B<--verbose>]
[I<ITEM>...]    

B<beam list> B<-h> | B<--help> | B<--usage> | B<--version>
    
=head1 DESCRIPTION

Lists existing backups.

=head1 OPTIONS

=over 4

=item B<-v>, B<--verbose>

Verbosely lists each backup item.

=back

Informational options:    

=over 4
    
=item B<-h>

Print a short help summary.

=item B<--help>

Display manpage.

=item B<--usage>

Print a short command line usage reminder.    

=back
    
=cut    

sub run {
    my $self = shift;
    my $verbose;
    
    GetOptionsFromArray(\@_,
			"verbose|v" => \$verbose)
	or exit(EX_USAGE);
			
    $self->abend(EX_USAGE, "bad number of arguments") if @_;
    
    my @records = $self->{history}->records;
    shift @records;
    my $i = 0;
    foreach my $rec (@records) {
	++$i;
	printf("%3d ", $i);
	print strftime "%c ", localtime $rec->start_time;
	print strftime "- %c ", localtime $rec->end_time;
	printf("% 3d% 3d% 3d  %s", $rec->cycle, $rec->round, $rec->level,
	       $rec->state_name);
	print "\n";
	if ($verbose) {
	    my %ent = $rec->items;
	    foreach my $item (split /\s+/, $self->get('core.items')) {
		if (exists($ent{$item})) {
		    printf("# %-24s %s %s\n", $item, $ent{$item}->state_name,
			   join(',', $ent{$item}->files));
		    delete $ent{$item};
		}
	    }
	    foreach my $item (sort keys %ent) {
		printf("# %-24s %s\n", $item, $ent{$item}->state_name);
	    }
	}		    
    }
}

1;
